#include <SDL.h>
#include <stdio.h>
#include <sys\types.h> 
#include <sys\stat.h>

#include "Constants.h"
#include "SnaLoader.h"
#include "Cpu.h"
#include "CpuUtilities.h"
#include "Io.h"
#include "Memory.h"
#include "Log.h"

void sna_load(char* filename) {
    log_write("Starting SNA load");
    log_write(filename);
    struct stat info;
    if (stat(filename, &info) != 0) {
        log_write("Error: SNA file not found");
        return;
    }

    if (info.st_size != 49179) {
        log_write("Error: SNA file must be exactly 49179 bytes long");
        return;
    }

    char* file = (Uint8*)malloc(info.st_size);
    if (file == NULL) {
        log_write("Error: unable to allocate memory for SNA file");
        return;
    }
    FILE* fp;
    errno_t result = fopen_s(&fp, filename, "rb");
    if (result) {
        log_write("Error: could not open SNA file");
        return;
    }
    // try to read a single block of info.st_size bytes
    size_t blocks_read = fread(file, info.st_size, 1, fp);
    if (blocks_read != 1) {
        log_write("Error: could not read from SNA file");
        return;
    }
    fclose(fp);

    // set up registers
    cpu_regs()->I = *((Uint8*)(file + 0x00));
    *cpu_regs()->HL_alt = *((Uint16*)(file + 0x01));
    *cpu_regs()->DE_alt = *((Uint16*)(file + 0x03));
    *cpu_regs()->BC_alt = *((Uint16*)(file + 0x05));
    *cpu_regs()->AF_alt = *((Uint16*)(file + 0x07));
    *cpu_regs()->HL = *((Uint16*)(file + 0x09));
    *cpu_regs()->DE = *((Uint16*)(file + 0x0B));
    *cpu_regs()->BC = *((Uint16*)(file + 0x0D));
    *cpu_regs()->IY = *((Uint16*)(file + 0x0F));
    *cpu_regs()->IX = *((Uint16*)(file + 0x11));

    cpu_regs()->IFF = 0;
    Uint8 bit0x13 = *((Uint8*)(file + 0x13));
    if (bit0x13 & 4) {
        cpu_regs()->IFF |= CPU_IFF2_BIT;
        // note: the SNA spec does not mention IFF1 because after loading the SNA
        //       execution is resumed by simulating a RETN instruction, which
        //       has the side effect of copying IFF2 into IFF1
    }

    cpu_regs()->R = *((Uint8*)(file + 0x14));
    *cpu_regs()->AF = *((Uint16*)(file + 0x15));
    cpu_regs()->SP = *((Uint16*)(file + 0x17));

    Uint8 interruptMode = *((Uint8*)(file + 0x19));
    switch (interruptMode) {
    case 0:
        cpu_set_interrupt_mode(Mode0);
        break;
    case 1:
        cpu_set_interrupt_mode(Mode1);
        break;
    case 2:
        cpu_set_interrupt_mode(Mode2);
        break;
    default:
        log_write("Error: SNA file must specify 0, 1, or 2 for interrupt mode (byte at offset 0x19)");
        free(file);
        return;
    }

    Uint8 borderColour = *((Uint8*)(file + 0x1A));
    io_write8_16bitaddr(PORT_BORDER_COLOUR, borderColour);

    memory_load(file + 0x1B, 16384, SPECTRUM_RAM_SIZE);
    free(file);
    log_write("Loaded SNA into ZX Spectrum memory");

    // simulate retn
    Uint16 branchTarget = cpu_pop16();
    cpu_regs()->PC = branchTarget;
    if (cpu_regs()->IFF & CPU_IFF2_BIT) cpu_regs()->IFF |= CPU_IFF1_BIT; else cpu_regs()->IFF &= ~CPU_IFF1_BIT;
    log_write("Performed a retn to resume SNA");
}