#include <math.h>
#include <string.h>
#include <stdio.h>
#include <SDL.h>
#include <SDL_audio.h>

#include "Constants.h"
#include "Log.h"

Uint16 _soundFrameDurationMs;

//const int SAMPLE_RATE_PER_SECOND = 44100;
//const int SAMPLE_RATE_PER_SECOND = 22050;
const int SAMPLE_RATE_PER_SECOND = 11025;
//const int SAMPLE_RATE_PER_SECOND = 8000;

const Sint16 AUDIO_HIGH_I = 0x3DF0;
const Sint16 AUDIO_LOW_I = 0x0000;

const float AUDIO_HIGH_F = 0.10f;
const float AUDIO_LOW_F = 0.0f;


static Uint8 _soundInitialized = 0;

//    ++-----------------------sample is signed if set
//    ||
//    ||       ++-----------sample is bigendian if set
//    ||       ||
//    ||       ||          ++---sample is float if set
//    ||       ||          ||
//    ||       ||          ||     +---sample bit size---+
//    ||       ||          ||     |                     |
//    15 14 13 12 11 10 09 08     07 06 05 04 03 02 01 00
SDL_AudioFormat _format;
Uint8 _formatBitsPerSample;
Uint8 _formatIsFloat;

static Uint16 _spectrumSpeakerBufferSize;
static Uint16 _spectrumSpeakerSamplingIntervalTstates;

// these two buffers collect samples from the ZX Spectrum speaker, alternating
// each video frame (50Hz) between being read or written
static Uint8* _spectrumSpeakerBuffer0 = NULL;
static Uint8* _spectrumSpeakerBuffer1 = NULL;

// these alternate between the two above buffers every frame
// as soon as one buffer has filled up, it becomes ready to be read by the SDL audio logic
// conversely, at the same time, the other buffer has been fully read, and becomes ready to collect samples
static Uint8** _spectrumSpeakerReadBufferPtr;
static Uint8** _spectrumSpeakerWriteBufferPtr;

// indexes used by the writer (50Hz interrupt) and reader (SDL's audio logic)
static Uint8 _spectrumSpeakerBufferReadPosition;
static Uint8 _spectrumSpeakerBufferWritePosition;

static volatile Uint8 _bufferSwapLock;
static Uint32 _tstatesAtLastSample;


// the read operation treats the buffer as linear, to not overwrite earlier samples
void _sound_spectrum_buffer_write(Uint8 value8) {
    if (!_soundInitialized) {
        return;
    }

    if (_spectrumSpeakerBufferWritePosition >= _spectrumSpeakerBufferSize) {
        // NOOP when buffer is full
        return;
    }

    (*_spectrumSpeakerWriteBufferPtr)[_spectrumSpeakerBufferWritePosition] = value8;
    _spectrumSpeakerBufferWritePosition++;
}

// the read operation treats the buffer as circular, to handle small overruns
Uint8 _sound_spectrum_buffer_read() {
    if (!_soundInitialized) {
        return 0xF3;
    }

    Uint8 value8 = (*_spectrumSpeakerReadBufferPtr)[_spectrumSpeakerBufferReadPosition];
    _spectrumSpeakerBufferReadPosition++;
    if (_spectrumSpeakerBufferReadPosition == _spectrumSpeakerBufferSize) {
        _spectrumSpeakerBufferReadPosition = 0;
    }
    return value8;
}

void _sound_buffer_filler_callback(void* consumerData, Uint8* sdlBuffer, int bytesToFill) {
    SDL_MemoryBarrierAcquire();
    _bufferSwapLock = 1;
    SDL_MemoryBarrierRelease();

    int lengthInSamples = bytesToFill / (_formatBitsPerSample/8);

    if (_formatIsFloat) {
        // samples must be floating point

        if (_formatBitsPerSample == 32) {
            float* buffer = (float*)sdlBuffer;
            for (int i = 0; i < lengthInSamples; i++) {
                float value = _sound_spectrum_buffer_read() ? AUDIO_HIGH_F : AUDIO_LOW_F;
                buffer[i] = value;
            }
        }
        else if (_formatBitsPerSample == 64) {
            double* buffer = (double*)sdlBuffer;
            for (int i = 0; i < lengthInSamples; i++) {
                double value = _sound_spectrum_buffer_read() ? (double)AUDIO_HIGH_F : (double)AUDIO_LOW_F;
                buffer[i] = value;
            }
        }
    }
    else {
        // samples must be integers

        if (_formatBitsPerSample == 16) {
            Uint16* buffer = (Uint16*)sdlBuffer;
            for (int i = 0; i < lengthInSamples; i++) {
                Uint16 value = _sound_spectrum_buffer_read() ? (Uint16)AUDIO_HIGH_I : (Uint16)AUDIO_LOW_I;
                buffer[i] = value;
            }
        }
        else if (_formatBitsPerSample == 32) {
            Uint32* buffer = (Uint32*)sdlBuffer;
            for (int i = 0; i < lengthInSamples; i++) {
                Uint32 value = _sound_spectrum_buffer_read() ? (Uint32)AUDIO_HIGH_I : (Uint32)AUDIO_LOW_I;
                buffer[i] = value;
            }
        }
    }
    
    SDL_MemoryBarrierAcquire();
    _bufferSwapLock = 0;
    SDL_MemoryBarrierRelease();
}

void sound_add_sample(Uint8 isSpeakerActive, Uint32 totalTStates) {
    if (!_soundInitialized) {
        return;
    }

    if (totalTStates - _tstatesAtLastSample >= _spectrumSpeakerSamplingIntervalTstates) {
        // this allows for a catchup on consecutive CPU instructions, if, for some reason,
        // we fall behind for more than one sampling interval
        _tstatesAtLastSample += _spectrumSpeakerSamplingIntervalTstates;
        _sound_spectrum_buffer_write(isSpeakerActive);
    }
}

void sound_notify_end_of_frame() {
    if (!_soundInitialized) {
        return;
    }

    // block until buffer feeder is done
    while (1) {
        SDL_MemoryBarrierAcquire();
        if (!_bufferSwapLock) {
            SDL_MemoryBarrierRelease();
            break;
        }
        SDL_MemoryBarrierRelease();
    }

    // swap buffers
    if (*_spectrumSpeakerReadBufferPtr == _spectrumSpeakerBuffer0) {
        *_spectrumSpeakerReadBufferPtr = _spectrumSpeakerBuffer1;
        *_spectrumSpeakerWriteBufferPtr = _spectrumSpeakerBuffer0;
    }
    else {
        *_spectrumSpeakerReadBufferPtr = _spectrumSpeakerBuffer0;
        *_spectrumSpeakerWriteBufferPtr = _spectrumSpeakerBuffer1;
    }
    _spectrumSpeakerBufferReadPosition = 0;
    _spectrumSpeakerBufferWritePosition = 0;
}

void sound_destroy() {
    if (!_soundInitialized) {
        return;
    }

    SDL_PauseAudio(1); // stop playing sound
    SDL_CloseAudio();

    if (_spectrumSpeakerBuffer0 != NULL) {
        free(_spectrumSpeakerBuffer0);
        _spectrumSpeakerBuffer0 = NULL;
    }

    if (_spectrumSpeakerBuffer1 != NULL) {
        free(_spectrumSpeakerBuffer1);
        _spectrumSpeakerBuffer1 = NULL;
    }

    if (_spectrumSpeakerReadBufferPtr != NULL) {
        free(_spectrumSpeakerReadBufferPtr);
        _spectrumSpeakerReadBufferPtr = NULL;
    }

    if (_spectrumSpeakerWriteBufferPtr != NULL) {
        free(_spectrumSpeakerWriteBufferPtr);
        _spectrumSpeakerWriteBufferPtr = NULL;
    }
}

void sound_start(Uint16 frameDurationMs) {
    sound_destroy();

    _soundFrameDurationMs = frameDurationMs;

    if (SDL_Init(SDL_INIT_AUDIO) != 0) {
        log_write("Error: could not initialize SDL audio");
        log_write((char*)SDL_GetError());
        return;
    }

    SDL_AudioSpec want;
    want.freq = SAMPLE_RATE_PER_SECOND;             // samples per second
    want.format = AUDIO_F32SYS;                     // sample type: 32bit float
    want.channels = 1;                              // mono
    want.samples = 16 * 1024;                       // buffer size
    want.callback = _sound_buffer_filler_callback;  // called by SDL to "feed" buffer
    want.userdata = NULL;                           // keeps track of current sample number
    want.silence = 0;

    char* message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Requesting audio spec %04x", want.format);
        log_write(message);
        free(message);
    }

    SDL_AudioSpec got;
    if (SDL_OpenAudio(&want, &got) != 0) {
        log_write("Error: could not open audio");
        log_write((char*)SDL_GetError());
        return;
    }
    _format = got.format;
    _formatBitsPerSample = SDL_AUDIO_BITSIZE(got.format);
    _formatIsFloat = SDL_AUDIO_ISFLOAT(got.format) ? 1 : 0;

    message = (char*)malloc(500 * sizeof(char));
    if (message) {
        sprintf_s(message, 490, "Got audio spec %04x", got.format);
        log_write(message);
        free(message);
    }

    if (SDL_AUDIO_ISBIGENDIAN(got.format)) {
        log_write("Warning: got audio spec that is big endian, which is unsupported; audio will likely be silence or garbage");
    }

    if (want.format != got.format) {
        log_write("Warning: got an audio spec different than the one requested");
    }

    // calculate ZX Spectrum speaker buffer size... it must be large enough
    // to hold enough samples to last as long as a video frame (20ms by default)
    const double bufferOversize = 1.095f; // experimental
    _spectrumSpeakerBufferSize = (Uint16)((double)SAMPLE_RATE_PER_SECOND * ((double)_soundFrameDurationMs / 1000.0f) * bufferOversize);
    // each sample is effectively 1bit (speaker active or inactive), so we store it as a single byte

    _spectrumSpeakerBuffer0 = (Uint8*)malloc(_spectrumSpeakerBufferSize * sizeof(Uint8));
    if (_spectrumSpeakerBuffer0 == NULL) {
        return;
    }
    _spectrumSpeakerBuffer1 = (Uint8*)malloc(_spectrumSpeakerBufferSize * sizeof(Uint8));
    if (_spectrumSpeakerBuffer1 == NULL) {
        return;
    }
    _spectrumSpeakerReadBufferPtr = (Uint8**)malloc(1 * sizeof(Uint8*));
    if (_spectrumSpeakerReadBufferPtr == NULL) {
        return;
    }
    _spectrumSpeakerWriteBufferPtr = (Uint8**)malloc(1 * sizeof(Uint8*));
    if (_spectrumSpeakerWriteBufferPtr == NULL) {
        return;
    }

    // zero out the two buffers
    memset(_spectrumSpeakerBuffer0, 0, _spectrumSpeakerBufferSize);
    memset(_spectrumSpeakerBuffer1, 0, _spectrumSpeakerBufferSize);

    // calculate sampling interval in tstates
    _spectrumSpeakerSamplingIntervalTstates = (Uint16)((double)TSTATES_PER_FRAME / (double)_spectrumSpeakerBufferSize);

    _spectrumSpeakerBufferReadPosition = 0;
    _spectrumSpeakerBufferWritePosition = 0;
    *_spectrumSpeakerReadBufferPtr = _spectrumSpeakerBuffer0;
    *_spectrumSpeakerWriteBufferPtr = _spectrumSpeakerBuffer1;

    _bufferSwapLock = 0;
    _tstatesAtLastSample = 0;

    SDL_PauseAudio(0);      // start playing sound
    _soundInitialized = 1;
}




// === BEGIN used for flat tone testing ===
/*
const double FREQUENCY = 200.0f;
static double PERIOD;
static double periodSamples;
static Uint16 sustainTime;
static Uint16 sampleCursor;
static Uint16 samplesRemainingAtCurrentLevel;
static Uint8 isLow;

// square wave amplitude
//const Sint16 AUDIO_HIGH = 0x3DF0;
//const Sint16 AUDIO_LOW = 0x0000;
*/
// === END used for flat tone testing ===



// DEBUG ROUTINES USED WITH FLAT TONE OUTPUT AT A SPECIFIED FREQUENCY

/*
void _sound_buffer_filler_callback__flat_square_tone(void* user_data, Uint8* raw_buffer, int bytes) {

    Sint16* buffer = (Sint16*)raw_buffer;
    int length = bytes / 2; // 2 bytes per sample for AUDIO_S16SYS

    for (int i = 0; i < length; i++, sampleCursor++) {
        if (!isLow) {
            buffer[i] = AUDIO_HIGH;
        }
        else {
            buffer[i] = AUDIO_LOW;
        }

        samplesRemainingAtCurrentLevel--;
        if (samplesRemainingAtCurrentLevel == 0) {
            isLow ^= 1;
            samplesRemainingAtCurrentLevel = sustainTime;
        }
    }
}

void sound_flat_tone_test() {
    PERIOD = 1.0f / FREQUENCY;
    periodSamples = SAMPLE_RATE_PER_SECOND * PERIOD;
    sustainTime = (Uint16)(periodSamples*2.0f);

    samplesRemainingAtCurrentLevel = sustainTime;
    sampleCursor = 0;
    isLow = 1;

    SDL_PauseAudio(0); // start playing sound
    SDL_Delay(5000); // wait while sound is playing
    SDL_PauseAudio(1); // stop playing sound

    sound_destroy();//DEBUG!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    while (1) {}
}
*/