#include <Windows.h>
#include <sys\timeb.h>
#include <SDL.h>
#include <stdio.h>

#include "Constants.h"
#include "Io.h"
#include "ULABus.h"
#include "Memory.h"
#include "Cpu.h"
#include "Log.h"
#include "Sound.h"

static SDL_Surface* _surface;
static SDL_Window* _window;

static BOOL _videoIsInitialized = FALSE;

HANDLE _callbackTimerHandle;
HANDLE _videoFrameMutex;

static volatile char _isDestroying;
static volatile char _callbackDestroyed;

struct timeb _videoStart, _videoEnd;

static Uint16 _videoFrameDurationMs;

static Uint32 _frames = 0;
static Uint32 _tstates = 0;
static Uint8 _zoom = 1;
static Uint32 _frameTStatesRemaining;
static Uint8 _isFlashInverse;


static Uint8 _palette[] = {
    0x00, 0x00, 0x00,   // black
    0x00, 0x22, 0xc7,   // blue
    0xd6, 0x28, 0x16,   // red
    0xd4, 0x33, 0xc7,   // magenta
    0x00, 0xc5, 0x25,   // green
    0x00, 0xc7, 0xc9,   // cyan
    0xcc, 0xc8, 0x2a,   // yellow
    0xca, 0xca, 0xca,   // white

    // bright versions
    0x00, 0x00, 0x00,   // black
    0x00, 0x2b, 0xfb,   // blue
    0xff, 0x33, 0x1c,   // red
    0xff, 0x40, 0xfc,   // magenta
    0x00, 0xf9, 0x2f,   // green
    0x00, 0xfb, 0xfe,   // cyan
    0xff, 0xfc, 0x36,   // yellow
    0xff, 0xff, 0xff,   // white
};

void _video_tstate_allocator_init() {
    _frameTStatesRemaining = TSTATES_PER_FRAME;
}

void _video_run_cpu_for_scanline() {
    if (_frameTStatesRemaining == 0) {
        return;
    }

    Uint32 allocation = SPECTRUM_SCAN_LINE_TSTATES;
    if (_frameTStatesRemaining < allocation) {
        allocation = _frameTStatesRemaining;
    }
    Uint32 tstatesConsumed = cpu_run_many(allocation);
    _tstates += tstatesConsumed;

    if (tstatesConsumed > _frameTStatesRemaining) {
        _frameTStatesRemaining = 0;
    }
    else {
        _frameTStatesRemaining -= tstatesConsumed;
    }
}

void _video_set_pixel(int x, int y, Uint8 r, Uint8 g, Uint8 b, Uint8 a) {
    Uint32* p_screen = (Uint32*)_surface->pixels;
    Uint32 hardwarePixelValue = SDL_MapRGBA(_surface->format, r, g, b, a);

    x -= HIDE_LEFT_BORDER_COUNT;
    y -= HIDE_TOP_BORDER_COUNT;

    int xLimit = VISIBLE_SPECTRUM_SCREEN_WIDTH;
    int yLimit = VISIBLE_SPECTRUM_SCREEN_HEIGHT;

    for (int yy = 0; yy < _zoom; yy++) {
        int offset = (_surface->pitch / sizeof(Uint32)) * (y*_zoom + yy) + (x*_zoom);
        if ((y*_zoom + yy) < 0 || (y * _zoom + yy) >= yLimit * _zoom) {
            continue;
        }
        for (int xx = 0; xx < _zoom; xx++) {
            if ((x * _zoom + xx) < 0 || (x * _zoom + xx) >= xLimit * _zoom) {
                continue;
            }
            *(p_screen + offset++) = hardwarePixelValue;
        }
    }
}

void _video_write_border_pixel(int x, int y) {
    Uint8 value = io_read8_border_colour() & 0x07;   // keep only colour

    Uint8 red = *(_palette + 3 * value + 0);
    Uint8 green = *(_palette + 3 * value + 1);
    Uint8 blue = *(_palette + 3 * value + 2);
    _video_set_pixel(x, y, red, green, blue, 255);
}

void _video_write_border_scan_line(int y) {
    for (int i = 0; i < SPECTRUM_SCREEN_WIDTH; ++i) {
        _video_write_border_pixel(i, y);
    }

    _video_run_cpu_for_scanline();
}

void _video_write_paper_scan_line(int y, int rowStartAddress) {
    int x = 0;

    // left border
    for (int i = 0; i < SPECTRUM_BORDER_WIDTH; i++) {
        _video_write_border_pixel(x + i, y);
    }
    x += SPECTRUM_BORDER_WIDTH;

    int attributesBaseAddress = 16384 + (SPECTRUM_PAPER_HEIGHT * SPECTRUM_PAPER_WIDTH) / 8;    // this is now at the start of attributes RAM
    int attributeRowNumber = (y - SPECTRUM_BORDER_LINES_BEFORE_PAPER) / 8;
    int attributeRowOffset = attributeRowNumber * (SPECTRUM_PAPER_WIDTH / 8);

    for (int i = 0; i < SPECTRUM_PAPER_WIDTH; i+=8) {
        Uint8 pixelByte = memory_read8(rowStartAddress + i/8);
        ula_bus_write8(pixelByte);
        
        int attributeByteOffset = attributeRowOffset + i / 8;

        int attributeAddress = attributesBaseAddress + attributeByteOffset;
        Uint8 attributeByte = memory_read8(attributeAddress);
        ula_bus_write8(attributeByte);

        // figure out R, G, B for the next 8 pixels
        // attributes
        // bits:       7        6     5   4   3      2   1   0
        //         FLASH   BRIGHT    P2  P1  P0     I2  I1  I0
        Uint8 isBright = attributeByte & 0x40;
        Uint8 isInverse = (attributeByte & 0x80) && _isFlashInverse;

        Uint8 inkColour = attributeByte & 0x07;
        Uint8 inkColourOffset = 3 * inkColour;
        if (isBright) {
            inkColourOffset = inkColourOffset + 3 * SPECTRUM_COLOUR_COUNT;
        }
        
        Uint8 inkRed = *(_palette + inkColourOffset + 0);
        Uint8 inkGreen = *(_palette + inkColourOffset + 1);
        Uint8 inkBlue = *(_palette + inkColourOffset + 2);

        Uint8 paperColour = (attributeByte >> 3) & 0x07;
        Uint8 paperColourOffset = 3 * paperColour;
        if (isBright) {
            paperColourOffset = paperColourOffset + 3 * SPECTRUM_COLOUR_COUNT;
        }

        Uint8 paperRed = *(_palette + paperColourOffset + 0);
        Uint8 paperGreen = *(_palette + paperColourOffset + 1);
        Uint8 paperBlue = *(_palette + paperColourOffset + 2);

        for (int j = 0; j < 8; j++) {
            if (!isInverse) {
                if (!(pixelByte & 0x80)) {
                    // paper colour
                    _video_set_pixel(x + i + j, y, paperRed, paperGreen, paperBlue, 255);
                }
                else {
                    // ink colour
                    _video_set_pixel(x + i + j, y, inkRed, inkGreen, inkBlue, 255);
                }
            }
            else {
                if (!(pixelByte & 0x80)) {
                    // ink colour
                    _video_set_pixel(x + i + j, y, inkRed, inkGreen, inkBlue, 255);
                }
                else {
                    // paper colour
                    _video_set_pixel(x + i + j, y, paperRed, paperGreen, paperBlue, 255);
                }
            }
            pixelByte = pixelByte << 1;
        }
    }
    x += SPECTRUM_PAPER_WIDTH;

    // right border
    for (int i = 0; i < SPECTRUM_BORDER_WIDTH; i++) {
        _video_write_border_pixel(x + i, y);
    }

    _video_run_cpu_for_scanline();
}

void _video_callback_worker() {
    if (_isDestroying) {
        _callbackDestroyed = 1;
        return;
    }

    _video_tstate_allocator_init();
    _isFlashInverse = _frames & 0x10;
    cpu_request_interrupt();

    SDL_LockSurface(_surface);

    int y = 0;

    // border before (above) paper
    for (int i = 0; i < SPECTRUM_BORDER_LINES_BEFORE_PAPER; i++) {
        _video_write_border_scan_line(y + i);
    }
    y += SPECTRUM_BORDER_LINES_BEFORE_PAPER;

    int rowStartAddressAtStartOfThird = 16384;

    // top third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);

    // middle third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24)*(SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);

    // bottom third of paper
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;

    ula_bus_go_inactive();

    // border after (below) paper
    for (int i = 0; i < SPECTRUM_BORDER_LINES_AFTER_PAPER; i++) {
        _video_write_border_scan_line(y + i);
    }

    SDL_UnlockSurface(_surface);
    SDL_UpdateWindowSurface(_window);

    _frames++;
    sound_notify_end_of_frame();
}

void __stdcall video_callback(PVOID lpParameter, BOOLEAN TimerOrWaitFired) {
    DWORD dwWaitResult = WaitForSingleObject(_videoFrameMutex, INFINITE);
    
    switch (dwWaitResult) {
        case WAIT_OBJECT_0:
            // this thread got ownership of the mutex
            __try {
                _video_callback_worker();
            }

            __finally {
                if (!ReleaseMutex(_videoFrameMutex))
                {
                    // an error happened
                }
            }
            break;

        case WAIT_ABANDONED:
            // this thread got ownership of an abandoned mutex
            return;
    }
}

Uint16 video_get_frame_duration_MS() {
    return _videoFrameDurationMs;
}


BOOL video_start(SDL_Window* window, char zoom, Uint16 framePeriodMs) {
    _window = window;
    _zoom = zoom;
    _videoFrameDurationMs = framePeriodMs;
    _frames = 0;
    _tstates = 0;

    _isDestroying = 0;
    _callbackDestroyed = 0;
    ftime(&_videoStart);

    _surface = SDL_GetWindowSurface(window);
    if (!_surface) {
        log_write("Error: unable to create SDL window surface");
        return FALSE;
    }
    
    _videoFrameMutex = CreateMutex(NULL, FALSE, NULL);
    if (!_videoFrameMutex) {
        log_write("Error: unable to create video frame timer mutex");
        return FALSE;
    }
    
    Uint16 framePeriod = video_get_frame_duration_MS();
    BOOL res = CreateTimerQueueTimer(&_callbackTimerHandle, NULL, video_callback, NULL, VIDEO_DELAY_TIMER_START_MS, framePeriod, WT_EXECUTEINTIMERTHREAD);
    if (!res) {
        log_write("Error: unable to start video frame timer");
        _callbackTimerHandle = NULL;
        return FALSE;
    }
    
    _videoIsInitialized = TRUE;
    log_write("Started video timer, yielding CPU control to it");
    return TRUE;
}

void video_destroy() {
    if (_callbackTimerHandle) {
        _isDestroying = 1;
        // we busy-wait for callback to become NOOP only if we actually initialized
        // the callback timer successfully
        while (!_callbackDestroyed) {}

        BOOL res = DeleteTimerQueueTimer(NULL, _callbackTimerHandle, NULL);
        if (!res) {
            log_write("Error: unable to delete video frame timer");
        }
    }

    ftime(&_videoEnd);

    if (_videoFrameMutex) {
        CloseHandle(_videoFrameMutex);
    }

    if (_surface) {
        SDL_FreeSurface(_surface);
    }

    // log statistics
    char* message = (char*)malloc(500 * sizeof(char));
    if (message) {
        long double tstates = (long double)_tstates;
        double frames = (double)_frames;

        sprintf_s(message, 490, "Total video frames: %d", _frames);
        log_write(message);

        Uint64 runTimeMs = (Uint64)(1000.0 * (_videoEnd.time - _videoStart.time)
            + (_videoEnd.millitm - _videoStart.millitm)) - (Uint32)VIDEO_DELAY_TIMER_START_MS;

        sprintf_s(message, 490, "Total duration: %I64u ms", runTimeMs);
        log_write(message);

        double fps = frames / ((double)runTimeMs/1000.f);
        sprintf_s(message, 490, "Average frames per second: %.3f", fps);
        log_write(message);

        double millisPerFrame = (double)runTimeMs / frames;
        Uint32 targetMillisPerFrame = video_get_frame_duration_MS();
        sprintf_s(message, 490, "Average milliseconds per frame: %.3f (target is %d)", millisPerFrame, targetMillisPerFrame);
        log_write(message);

        sprintf_s(message, 490, "Average tstates per video frame: %.3f (target is %d)", tstates/frames, SPECTRUM_VIDEO_FRAME_TSTATES);
        log_write(message);
        free(message);
    }
}
